# -*- coding: utf-8 -*-

from akagi.iterators import csv_iterator
from akagi.iterators import binary_iterator

from akagi.iterators.csv_iterator import CSVIterator
from akagi.iterators.binary_iterator import BinaryIterator



import re
import os
from six.moves.urllib import parse, request

from akagi.content import Content
from akagi.iterator import Iterator


class URLContent(Content):
    is_local = False

    def __init__(self, urlstr, file_format='csv'):
        self._url = parse.urlparse(urlstr)
        self.file_format = file_format
        self.iterator_class = Iterator.get_iterator_class(file_format)

    def __iter__(self):
        return self.iterator_class(self)

    @property
    def key(self):
        return os.path.join(self._url.netloc, re.sub(r'^/', '', self._url.path))

    @property
    def _body(self):
        return request.urlopen(self._url.geturl())



class BinaryIterator(object):
    def __init__(self, content):
        self.body = content.raw_body
        self._stop = False

    @classmethod
    def open_file(self, path):
        return open(path, 'rb')

    def __next__(self):
        if self._stop:
            raise StopIteration
        else:
            self._stop = True
            return self.body



import os
from six import BytesIO

import akagi
from akagi.content import Content
from akagi.iterator import Iterator


class S3Content(Content):
    is_local = False

    def __init__(self, bucket_name, key, file_format='csv'):
        self._bucket_name = bucket_name
        self._key = key
        self.file_format = file_format
        self.iterator_class = Iterator.get_iterator_class(file_format)

    def __iter__(self):
        return self.iterator_class(self)

    @property
    def key(self):
        return os.path.join(self._bucket_name, self._key)

    @property
    def _body(self):
        out_io = BytesIO()
        akagi.get_resource('s3').Object(self._bucket_name, self._key).download_fileobj(out_io)
        out_io.seek(0)

        return out_io



import os
import akagi

from akagi.log import logger

from akagi.contents import S3Content, LocalFileContent, SpreadsheetContent, URLContent


def data_files_for_s3_prefix(bucket_name, prefix, file_format='csv'):
    return [DataFile.s3(bucket_name, obj.key, file_format) for obj in
            akagi.get_resource('s3').Bucket(bucket_name).objects.filter(Prefix=prefix)]


def data_files_for_s3_keys(bucket_name, keys, file_format='csv'):
    return [DataFile.s3(bucket_name, k, file_format) for k in keys]


def data_files_for_dir(dir_path, file_format='csv'):
    return [DataFile.local_file(os.path.join(dir_path, p), file_format) for p in
            os.listdir(dir_path)]


def data_files_for_urls(urls, file_format='csv'):
    return [DataFile.url(url, file_format) for url in urls]


class DataFile(object):
    def __init__(self, content):
        self._content = content

    def __iter__(self):
        return iter(self.content)

    @classmethod
    def s3(cls, bucket, key, file_format='csv'):
        content = S3Content(bucket, key, file_format)
        return DataFile(content)

    @classmethod
    def url(cls, url, file_format='csv'):
        content = URLContent(url, file_format)
        return DataFile(content)

    @classmethod
    def local_file(cls, path, file_format='csv'):
        content = LocalFileContent(path, file_format)
        return DataFile(content)

    @classmethod
    def spreadsheet(cls, sheet_id, sheet_range='A:Z'):
        content = SpreadsheetContent(sheet_id, sheet_range)
        return DataFile(content)

    def filename(self):
        '''Retrieve filename'''

    @property
    def path(self):
        return os.path.join(akagi.home(), 'cache', self._content.key)

    def _cache_content(self):
        '''Cache its content and return its path on local file system.'''

        dirname = os.path.dirname(self.path)

        try:
            os.makedirs(dirname)
        except OSError:
            if not os.path.isdir(dirname):
                raise

        if os.path.isfile(self.path) or os.path.isdir(self.path):
            logger.debug("Skipped %(key)s" % ({"key": self._content.key}))
        else:
            with open(self.path, 'wb') as f:
                f.write(self._content.raw_body.read())
                logger.debug("Saved %(key)s to %(path)s" % ({"key": self._content.key, "path": self.path}))

        return self.path

    @property
    def content(self):
        if not self._content.is_local:
            self._content = LocalFileContent(self._cache_content(), file_format=self._content.file_format)

        return self._content



import json
import os

import itertools
import abc
import akagi
from akagi.log import logger

from concurrent.futures import ThreadPoolExecutor
import six


@six.add_metaclass(abc.ABCMeta)
class DataSource(object):
    '''DataSource is an abstract base class of all data sources
    '''

    def save(self, force=False):
        paths = []
        paths.append(self._save_metadata())

        with ThreadPoolExecutor() as executor:
            paths = list(executor.map(self._save_data_file, self.bundle.data_files))

        return paths

    @abc.abstractproperty
    def data_files(self):
        '''The data files associated to the bundle.'''

    def _save_data_file(self, data_file):
        path = os.path.join(self._local_cache_dir, data_file.key)
        dirname = os.path.dirname(path)

        if not os.path.isdir(dirname):
            try:
                os.makedirs(dirname)
            except FileExistsError:
                pass

        if os.path.isfile(path) or os.path.isdir(path):
            logger.debug("Skipped %(key)s" % ({"key": data_file.key}))
        else:
            with open(path, 'wb') as f:
                f.write(data_file.raw_content)
                logger.debug("Saved %(key)s to %(path)s" % ({"key": data_file.key, "path": path}))

        return path

    def _save_metadata(self):
        path = self._local_cache_dir + '.json'

        metadata = {'iterator_class': self.bundle.iterator_class.__name__}

        with open(path, 'w') as f:
            json.dump(metadata, f)

        return path

    def exists_local(self):
        # TODO: Prone to error during self.save()
        return len(os.listdir(self._local_cache_dir)) > 0

    @property
    def _local_cache_dir(self):
        path = os.getenv('AKAGI_LOCAL_CACHE', akagi.home())
        if not os.path.isdir(path):
            os.makedirs(path, mode=0o755)

        cache_dir = os.path.join(path, 'cache', self._hex_hash)

        if not os.path.isdir(cache_dir):
            os.makedirs(cache_dir, mode=0o755)

        return cache_dir

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    def __iter__(self):
        return itertools.chain(*self.data_files)



import hashlib
import io
import six
import abc

import filetype

from filetype.types.archive import Gz
from akagi.utils import gzip_decompress


@six.add_metaclass(abc.ABCMeta)
class Content(object):
    sig_length = 256

    @property
    def body(self):
        if isinstance(self.sniff(), Gz):
            return gzip_decompress(self._body)
        else:
            return self._body

    @property
    def raw_body(self):
        return self._body

    @property
    def decoded_body(self):
        if six.PY2:
            return self.body
        else:
            if isinstance(self.body, io.TextIOBase):
                return self.body
            else:
                return six.StringIO(self.body.read().decode('utf-8'))

    @abc.abstractproperty
    def _body(self):
        '''Body of the content.'''

    def sniff(self):
        self._body.seek(0)

        if six.PY2:
            sig = bytearray(self.raw_body.read(self.sig_length))
        else:
            sig = self.raw_body.read(self.sig_length)

        self._body.seek(0)

        return filetype.guess(sig)

    @abc.abstractproperty
    def key(self):
        '''Retrieve key for the content'''



import hashlib
from datetime import datetime
import sys
import boto3
import re
import six
import os

import psycopg2

import akagi
from akagi.data_source import DataSource
from akagi.log import logger
from akagi.utils import normalize_path
from akagi.data_file import data_files_for_s3_prefix


class RedshiftDataSource(DataSource):
    '''RedshiftDataSource replesents a set of row data as a result of query param.
    It uses UNLOAD command and intermediate Amazon S3 bucket.
    '''
    def __init__(self, query_str, bucket_name=None, db_conf={}, no_cache=False):
        self._query_str = query_str
        self.__bucket_name = bucket_name
        self.__db_conf = db_conf
        self.__pgpass = None
        self._no_cache = no_cache
        self._data_files = None

    @property
    def _exists_on_s3(self):
        for _ in akagi.get_resource('s3').Bucket(self._bucket_name)\
                .objects.filter(Prefix=self._unload_root_prefix).limit(1):
            return True

        return False

    @property
    def _unload_root_prefix(self):
        p = os.path.join(
            os.getenv('AKAGI_UNLOAD_PREFIX', 'akagi_unload'),
            self._hex_hash)

        if p.endswith('/'):
            return p
        else:
            return p + '/'

    @property
    def data_files(self):
        if self._data_files is None:
            if not self._exists_on_s3:
                unload_prefix = self._unload_root_prefix
                datetime.utcnow().strftime("%Y%m%d_%H%M%f")  # XXX: utcnow()?

                query = UnloadQuery(self._query_str, self._bucket_name, unload_prefix)

                logger.debug("Executing query on Redshift")
                logger.debug("\n" + query.body + "\n")  # avoid logging unload query since it has raw credentials inside
                self._cursor.execute(query.sql)
                logger.debug("Finished")

            self._data_files = data_files_for_s3_prefix(self._bucket_name, self._latest_prefix_on_s3)

        return self._data_files

    @property
    def _latest_prefix_on_s3(self):
        prefixes = list(akagi.get_resource('s3').Bucket(self._bucket_name).objects.filter(Prefix=self._unload_root_prefix))

        if len(prefixes) > 0:
            return prefixes[-1].key
        else:
            None

    @property
    def _connection(self):
        connection = psycopg2.connect(**self._db_conf)
        connection.autocommit = 'REDSHIFT_DISABLE_AUTOCOMMIT' not in os.environ
        return connection

    @property
    def _cursor(self):
        return self._connection.cursor()

    @property
    def _db_conf(self):
        return {
            'host':
            self.__db_conf.get('host') or os.getenv('REDSHIFT_DB_HOST', self._pgpass['db_host']),
            'user':
            self.__db_conf.get('user') or os.getenv('REDSHIFT_DB_USER', self._pgpass['db_user']),
            'dbname':
            self.__db_conf.get('dbname') or os.getenv('REDSHIFT_DB_NAME', self._pgpass['db_name']),
            'password':
            self.__db_conf.get('password') or os.getenv('REDSHIFT_DB_PASS', self._pgpass['db_pass']),
            'port':
            self.__db_conf.get('port') or os.getenv('REDSHIFT_DB_PORT', self._pgpass['db_port'])}

    @property
    def _pgpass(self):
        if self.__pgpass is None:
            self.__pgpass = {}

            with open(os.path.expanduser(os.path.join('~', '.pgpass'))) as f:
                args = [s.strip() for s in f.read().split(':')]

                if len(args) == 5:
                    (
                        self.__pgpass['db_host'], self.__pgpass['db_port'],
                        self.__pgpass['db_name'], self.__pgpass['db_user'],
                        self.__pgpass['db_pass']) = args

        return self.__pgpass

    def __exit__(self, *exc):
        self.bundle.clear()
        return False

    @property
    def _hex_hash(self):
        return hashlib.sha256(self._query_str.encode('utf-8')).hexdigest()

    @property
    def _bucket_name(self):
        if self.__bucket_name is None:
            try:
                self.__bucket_name = os.environ['AKAGI_UNLOAD_BUCKET']
            except KeyError:
                logger.error('Environment variable AKAGI_UNLOAD_BUCKET must be set when using RedshiftDataSource.')
                sys.exit(1)

        return self.__bucket_name


class Query(object):
    def __init__(self, body):
        self._body = body

    def wrap(self, query):
        raise NotImplementedError

    @property
    def body(self):
        if six.PY2:
            return self._body.decode('utf-8')
        else:
            return self._body

    def __str__(self):
        return self.body


class UnloadQuery(Query):
    def __init__(self, body, bucket_name, prefix, sort=False):
        super(UnloadQuery, self).__init__(body)

        self._bucket_name = bucket_name
        self._prefix = prefix
        self._sort = sort

    @property
    def sql(self):
        s3_url = "%(bucket_name)s/%(prefix)s" % ({
            'bucket_name': self._bucket_name,
            'prefix': self._prefix})

        s3_url = "s3://" + normalize_path(s3_url)

        return """
unload ('%(query)s')
to '%(s3_url)s'
credentials '%(credential_string)s'
gzip
allowoverwrite
parallel %(enable_sort)s
delimiter ',' escape addquotes
            """ % ({
            'query': re.sub(r"'", "\\\\'", self.body),
            's3_url': s3_url,
            'credential_string': self._credential_string,
            'enable_sort': "on" if self._sort else "off"})

    @property
    def _credential_string(self):
        credentials = []

        if self._credential is not None:
            if self._credential.access_key:
                credentials.append("aws_access_key_id=%s" % (self._credential.access_key))

            if self._credential.secret_key:
                credentials.append("aws_secret_access_key=%s" % (self._credential.secret_key))

            if self._credential.token:
                credentials.append("token=%s" % (self._credential.token))

        return ';'.join(credentials)

    @property
    def _credential(self):
        session = boto3.session.Session()
        return session.get_credentials()



from abc import ABCMeta, abstractmethod
from akagi.iterators import CSVIterator, BinaryIterator
import six
import collections


@six.add_metaclass(ABCMeta)
class Iterator(collections.abc.Iterator):
    @abstractmethod
    def decode(self, content):
        '''Return decoded content.'''

    @classmethod
    def get_iterator_class(cls, file_format):
        if file_format.lower() == 'csv':
            return CSVIterator
        elif file_format.lower() == 'binary':
            return BinaryIterator
        else:
            raise Exception("Unsupported file format %(file_format)s." % locals())



from akagi.data_source import DataSource

from akagi.data_file import data_files_for_s3_prefix, data_files_for_s3_keys


class S3DataSource(DataSource):
    '''S3DataSource replesents a set of files on Amazon S3 bucket.
    '''

    @classmethod
    def for_prefix(cls, bucket_name, prefix, file_format='binary', no_cache=False):
        return S3DataSource(bucket_name, prefix=prefix, keys=None, file_format=file_format, no_cache=no_cache)

    @classmethod
    def for_keys(cls, bucket_name, keys, file_format='binary', no_cache=False):
        return S3DataSource(bucket_name, prefix=None, keys=keys, file_format=file_format, no_cache=no_cache)

    @classmethod
    def for_key(cls, bucket_name, key, file_format='binary', no_cache=False):
        return S3DataSource(bucket_name, prefix=None, keys=[key], file_format=file_format, no_cache=no_cache)

    def __init__(self, bucket_name, prefix=None, keys=None, file_format='binary', no_cache=False):
        self._file_format = file_format
        self._bucket_name = bucket_name
        self._keys = keys
        self._prefix = prefix
        self._no_cache = no_cache
        self._data_files = None

    @property
    def data_files(self):
        if self._data_files is None:
            if self._prefix is not None:
                self._data_files = data_files_for_s3_prefix(self._bucket_name, self._prefix, self._file_format)
            elif self._keys is not None:
                self._data_files = data_files_for_s3_keys(self._bucket_name, self._keys, self._file_format)
        return self._data_files



from akagi.data_source import DataSource

from akagi.data_file import data_files_for_urls


class URLDataSource(DataSource):
    '''URLDataSource replesents a set of files on remote location.
    '''

    @classmethod
    def for_urls(cls, urls, file_format='csv', no_cache=False):
        return URLDataSource(urls, file_format=file_format, no_cache=no_cache)

    def __init__(self, urls, file_format='csv', no_cache=False):
        self._urls = urls
        self._file_format = file_format

    @property
    def data_files(self):
        return data_files_for_urls(self._urls, self._file_format)



from akagi.data_source import DataSource
from akagi.data_file import DataFile


class SpreadsheetDataSource(DataSource):
    '''SpreadsheetSource replesents a data on Google Spreadsheets
    '''

    def __init__(self, sheet_id, sheet_range='A:Z', no_cache=False):
        self._sheet_id = sheet_id
        self._sheet_range = sheet_range

    @property
    def data_files(self):
        return [DataFile.spreadsheet(self._sheet_id, self._sheet_range)]



from __future__ import absolute_import

# -*- coding: utf-8 -*-
__author__ = """Yuichiro Someya"""
__email__ = 'ayemos.y@gmail.com'
__version__ = '0.3.0'


import os
import boto3

from akagi import data_file

from akagi import iterators

from akagi import content
from akagi import contents


def home():
    d = os.getenv('AKAGI_HOME', os.path.expanduser('~/.akagi'))

    try:
        os.makedirs(d)
    except OSError:
        if not os.path.isdir(d):
            raise

    return d


_s3 = None


def get_resource(name):
    if name == 's3':
        global _s3

        if _s3 is None:
            _s3 = boto3.resource('s3')

        return _s3
    else:
        raise Exception("Invalid resouce name %(name)s" % locals())



import sys
import os
import logging

logger = logging.getLogger('akagi')
logger.setLevel(logging.INFO)


sh = logging.StreamHandler(sys.stdout)
sh.setLevel(logging.INFO)

log_file_path = os.path.join('/', 'tmp', 'akagi.log')

fh = logging.FileHandler(log_file_path)
fh.setLevel(logging.DEBUG)  # log all

logger.addHandler(sh)
logger.addHandler(fh)
logger.debug("Logged to %(log_file_path)s" % locals())



import csv
import os
import json

from six import StringIO, BytesIO

from googleapiclient import discovery
from google.oauth2 import service_account

from akagi.content import Content
from akagi.iterator import Iterator


SCOPES = ['https://www.googleapis.com/auth/spreadsheets.readonly']


class SpreadsheetContent(Content):
    is_local = False

    def __init__(self, sheet_id, sheet_range='A:Z'):
        self._sheet_id = sheet_id
        self._range = sheet_range
        self.__credentials = None
        self.file_format = 'csv'
        self.iterator_class = Iterator.get_iterator_class(self.file_format)

    def __iter__(self):
        return self.iterator_class(self)

    @property
    def key(self):
        return 'spredsheet_' + self._sheet_id

    # XXX: Workaround for array-like result from Spreadsheet API
    @property
    def _body(self):
        result = self._discovered.spreadsheets().values()\
            .get(spreadsheetId=self._sheet_id, range=self._range).execute()
        out_io = StringIO()
        writer = csv.writer(out_io, delimiter=',')
        writer.writerows(result.get('values', []))
        out_io.seek(0)

        return BytesIO(out_io.read().encode())

    @property
    def _service_account_info(self):
        credential_path = os.getenv('GOOGLE_APPLICATION_CREDENTIALS', None)
        if credential_path is not None:
            with open(credential_path, 'r') as f:
                return json.load(f)
        else:
            raise Exception('Service account credentials must be specified by '
                            'GOOGLE_APPLICATION_CREDENTIALS for Spreadsheet support.')

    @property
    def _credentials(self):
        if self.__credentials is None:
            self.__credentials = service_account.Credentials.from_service_account_info(
                self._service_account_info, scopes=SCOPES)

        return self.__credentials

    @property
    def _discovered(self):
        return discovery.build('sheets', 'v4', credentials=self._credentials)



from akagi.data_source import DataSource
from akagi.data_file import data_files_for_dir


class LocalDataSource(DataSource):
    '''LocalDataSource replesents a set of files on local file system.
    '''

    def __init__(self, path, file_format='csv'):
        self._path = path
        self._file_format = file_format

    def __exit__(self, *exc):
        return False

    @property
    def data_files(self):
        return data_files_for_dir(self._path, self._file_format)



import re
import six
from six import BytesIO

import gzip


def gzip_decompress(data):
    if six.PY2:
        in_io = BytesIO()
        in_io.write(data.read())
        in_io.seek(0)
        return BytesIO(gzip.GzipFile(fileobj=in_io, mode='rb').read())
    else:
        return BytesIO(gzip.decompress(data.read()))


def normalize_path(path):
    return path and re.sub(r'^/', '', re.sub(r'\/{2,}', '/', path))



import csv


class CSVIterator(object):
    def __init__(self, content, skip_errors=True, delimiter=','):
        self.body = content.decoded_body
        self._skip_errors = skip_errors

        self._iterator = csv.reader(self.body, escapechar='\\', delimiter=delimiter)

    @classmethod
    def open_file(cls, path):
        return open(path, newline='')

    def next(self):
        try:
            return next(self._iterator)
        except StopIteration as e:
            raise e
        except Exception as e:
            if self._skip_errors:
                return next(self._iterator)
            else:
                raise e
    __next__ = next



# -*- coding: utf-8 -*-

from akagi.data_sources import redshift_data_source
from akagi.data_sources import s3_data_source
from akagi.data_sources import local_data_source
from akagi.data_sources import spreadsheet_data_source
from akagi.data_sources import url_data_source


from akagi.data_sources.redshift_data_source import RedshiftDataSource
from akagi.data_sources.s3_data_source import S3DataSource
from akagi.data_sources.local_data_source import LocalDataSource
from akagi.data_sources.spreadsheet_data_source import SpreadsheetDataSource
from akagi.data_sources.url_data_source import URLDataSource



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    'boto3>=1.4.4',
    'psycopg2>=2.7',
    'filetype>=1.0.0',
    'google-api-python-client>=1.6.3',
    'google-auth>=1.0.2'
]

test_requirements = [
    'codeclimate-test-reporter>=0.1.2'
]

setup(
    name='akagi',
    version='0.3.0',
    description="Codenize your data sources",
    long_description=readme,
    author="Yuichiro Someya",
    author_email='ayemos.y@gmail.com',
    url='https://github.com/ayemos/akagi',
    packages=[
        'akagi',
        'akagi.data_sources',
        'akagi.contents',
        'akagi.iterators'
    ],
    package_dir={'akagi':
                 'akagi'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='akagi',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



from akagi.content import Content
from akagi.iterator import Iterator


class LocalFileContent(Content):
    is_local = True

    def __init__(self, path, file_format='csv'):
        self.path = path
        self.file_format = file_format
        self.iterator_class = Iterator.get_iterator_class(file_format)

    def __iter__(self):
        return self.iterator_class(self)

    @property
    def key(self):
        return self.path

    @property
    def _body(self):
        return open(self.path, 'rb')



# -*- coding: utf-8 -*-

from akagi.contents import s3_content
from akagi.contents import local_file_content
from akagi.contents import spreadsheet_content
from akagi.contents import url_content

from akagi.contents.s3_content import S3Content
from akagi.contents.local_file_content import LocalFileContent
from akagi.contents.spreadsheet_content import SpreadsheetContent
from akagi.contents.url_content import URLContent
